# Copyright (c) 2025 Yoann Piétri
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

"""
A library for laser safety.
"""

__version__ = "0.9.0"

DISCLAIMER = """This code is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

In particular, you should ALWAYS check the results provided by this code before using them for
actual laser safety.
"""
